#include <netinet/in.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "response.h"
#include "scan.h"
#include "thread.h"

pthread_mutex_t g_getservent;

void *routine(void *p_data)
{
	struct scan *thread_data = p_data;
	struct scan scan_data;
	memcpy(&scan_data, thread_data, sizeof(struct scan));

	for (uint16_t port = thread_data->port_start;
	     port <= thread_data->port_end; port++) {
		scan_data.port_start = port;
		scan_data.port_end = port;
		scan_data.responses =
		    &thread_data->responses[port - thread_data->port_start];
		if (scan(&scan_data))
			return NULL;
	}

	return NULL;
}

static struct scan *init_threads_data(const struct scan *general,
                                      uint8_t nb_threads)
{
	struct scan *threads_data = malloc(sizeof(struct scan) * nb_threads);
	if (!threads_data) {
		dprintf(2, "ft_nmap: allocation of threads failed\n");
		return NULL;
	}

	const uint16_t port_range = general->port_end - general->port_start + 1;
	const uint16_t ports_per_thread = port_range / nb_threads;
	uint16_t remaining_ports = port_range % nb_threads;
	uint16_t port_start = general->port_start;
	for (uint8_t i = 0; i < nb_threads; i++) {
		memcpy(&threads_data[i], general, sizeof(struct scan));
		threads_data[i].port_start = port_start + i * ports_per_thread;
		threads_data[i].port_end = (port_start - 1) +
		                           (i + 1) * ports_per_thread +
		                           (remaining_ports ? 1 : 0);
		threads_data[i].responses =
		    general->responses +
		    (threads_data[i].port_start - general->port_start);
		if (remaining_ports) {
			remaining_ports--;
			port_start++;
		}
	}

	return threads_data;
}

int create_threads(struct scan *general, uint8_t nb_threads)
{
	if (nb_threads == 0) {
		routine(general);
		return 0;
	}

	struct scan *threads_data = init_threads_data(general, nb_threads);
	if (threads_data == NULL)
		return -1;
	pthread_t *threads = malloc(sizeof(pthread_t) * nb_threads);
	if (!threads) {
		free(threads_data);
		return -1;
	}

	pthread_mutex_init(&g_getservent, NULL);

	for (uint8_t i = 0; i < nb_threads; i++) {
		if (pthread_create(&threads[i], NULL, routine,
		                   &threads_data[i])) {
			dprintf(2, "ft_nmap: error during pthread_create()\n");
			return -1;
		}
	}

	for (uint8_t i = 0; i < nb_threads; i++) {
		if (pthread_join(threads[i], NULL)) {
			dprintf(2, "ft_nmap: error during pthread_join()\n");
			return -1;
		}
	}
	free(threads_data);
	free(threads);

	return 0;
}
